use anyhow::{Result, bail};
use clap::{Parser, Subcommand};
use std::env;
use tokio::process::Command;
use reqwest::{Client, ClientBuilder};

pub mod built_info {
    include!(concat!(env!("OUT_DIR"), "/built.rs"));
}

fn reqwests_version() -> Option<String> {
    for (name,version) in built_info::DEPENDENCIES {
        if name == "reqwest" {
            return Some(version.to_string());
        }
    }
    None
}

#[derive(Parser)]
#[command(about = "Go updater")]
struct Cli {
    #[command(subcommand)]
    command: Option<Commands>,
}

#[derive(Subcommand)]
enum Commands {
    /// Update go
    Update {},
    Info {},
}

/// Query `go version`
async fn go_version() -> Result<String> {
    let output = Command::new("go").arg("version").output().await?;
    if output.status.success() {
        // Ok! We have something!
        return Ok(String::from_utf8_lossy(output.stdout.as_slice()).to_string());
    }
    bail!("Failed to query go version.");
}

// or possibly pathsearch or search_path

async fn go_where() -> Result<String> {
    let output = Command::new("which").arg("go").output().await?;
    if output.status.success() {
        return Ok(String::from_utf8_lossy(output.stdout.as_slice()).to_string());
    }
    bail!("Failed to locate go binary.");
}
const GO_URL: &str = "https://go.dev/dl/";
// 2 MB download of html...

static APP_USER_AGENT : &str = concat!(
    env!("CARGO_PKG_NAME"),
    "/",
    env!("CARGO_PKG_VERSION"),
);

/*
CSS Selector elements of interest:  (a href="" part).
href are relative...

div[class="downloadWrapper"] > a[class="download downloadBox"]

<a class="download downloadBox" href="/dl/go1.24.0.windows-amd64.msi">
<a class="download downloadBox" href="/dl/go1.24.0.linux-amd64.tar.gz">

table[class="downloadtable"] > tr[class=" "] > td[class="filename"] > a[class="download"]

Or possibly,

table[class="downloadtable"] > tr[class=" "]
or tr[class="highlight "] ?
and grab the section of td's.  class=filename has a href, last has SHA256.

<a class="download" href="/dl/go1.24.0.src.tar.gz">go1.24.0.src.tar.gz</a>
<a class="download" href="/dl/go1.24.0.linux-amd64.tar.gz">go1.24.0.linux-amd64.tar.gz</a>
<a class="download" href="/dl/go1.24.0.linux-arm64.tar.gz">go1.24.0.linux-arm64.tar.gz</a>
<a class="download" href="/dl/go1.24.0.windows-amd64.zip">go1.24.0.windows-amd64.zip</a>
<a class="download" href="/dl/go1.24.0.windows-amd64.msi">go1.24.0.windows-amd64.msi</a>

 */

async fn get_go_downloads() -> Result<()> {
    /* 

    // This returns an error.  We only get the headers on a valid, successful
    // request.  There doesn't seem to be a sane way to get the reqwest default
    // useragent string...
    
    let reqwest_default = reqwest::Client::new();
    let resp = reqwest_default.get("http://127.0.0.254:2888").send().await;
    println!("{:?}", resp); // .headers());

    */
    
    let client = reqwest::Client::builder()
    .user_agent(APP_USER_AGENT)
    .build()?;
    let resp = client.get(GO_URL).send().await?;
    println!("{:?}", resp);
    // Response headers, not the request headers...
    // println!("{:?}", resp.headers());
    println!("{}", resp.text().await?);
    Ok(())
}


#[tokio::main]
async fn main() -> Result<()> {
    let cli = Cli::parse();

    // Get go environment
    let go_path = env::var("GOPATH").unwrap_or(String::new());
    let go_root = env::var("GOROOT").unwrap_or(String::new());
    let go_version = go_version().await;
    let go_where = go_where().await;

    // println!("{:?}", built_info::DEPENDENCIES);
    println!("Reqwest version: {:?}", reqwests_version());
    
    println!("path {} / root {}", go_path, go_root);
    println!("version: {:?}", go_version);
    println!("where: {:?}", go_where);
    // println!("Result: {:?}", get_go_downloads().await );

    // Get go version and path

    match &cli.command {
        Some(Commands::Update {}) => {}
        Some(Commands::Info {}) => {}
        None => {
            // Display help.
            let _show_help: Cli = Cli::parse_from(["--help"]);
        }
    }
    Ok(())
}
